from .elements import Unit, Option, Array, Structure, Variant, Map, List, Element
from .values import Boolean, Unsigned8, Unsigned16, Unsigned32, Unsigned64, Unsigned128, Signed8, Signed16, Signed32, \
    Signed64, Signed128, Char, String, Bytes, Value

from . import length
from . import ser as __ser__
from . import de as __de__
from .decode import decode as __decode__


def encode(s: Element) -> bytes:
    """
    Encode an Element object into tycho bytes.
    :param s:
    :return:
    """
    return s.encode()


def decode(s: bytes) -> Element:
    """
    Decode tycho bytes into Element object(s)
    :param s:
    :return:
    """
    return __decode__(s)


def serialise(s: dict or list or bool or int or float or str or bytes or None) -> Element:
    """
    Serialise a python object into a Element object(s)
    :param s:
    :return:
    """
    return __ser__.serialise(s)


def deserialise(s: Element) -> dict or list or bool or int or float or str or bytes or None:
    """
    Deserialise a Element object into python object(s)
    :param s:
    :return:
    """
    return __de__.deserialise(s)


def to_bytes(s: dict or list or bool or int or float or str or bytes or None) -> bytes:
    """
    Encode a python object into tycho bytes.

    Same as: `encode(serialise(object))`
    :param s:
    :return:
    """
    return encode(serialise(s))


def from_bytes(s: bytes) -> dict or list or bool or int or float or str or bytes or None:
    """
    Decode tycho bytes into a python object.

    Same as: `deserialise(decode(object))`
    :param s:
    :return:
    """
    return deserialise(decode(s))


__tycho__ = "0.4.0"



from io import BytesIO


def encode_length(n: int):
    if n < 0:
        raise ValueError("Length is less than 0")
    elif n < 0x3F:
        return (n & 0x3F).to_bytes(1, "big")
    elif n < 0x3FFF:
        return ((n & 0x3FFF) + 0x4000).to_bytes(2, "big")
    elif n < 0x3FFFFF:
        return ((n & 0x3FFFFF) + 0x800000).to_bytes(3, "big")
    elif n < 0x3FFFFFFF:
        return ((n & 0x3FFFFFFF) + 0xc0000000).to_bytes(4, "big")
    else:
        raise ValueError("Length is larger than (2^30)-1")


def decode_length(r: BytesIO):
    leading = r.read(1)[0]
    size = int(leading) >> 6
    read = bytes([leading & 0x3F]) + r.read(size)
    return int.from_bytes(read, "big")




from __future__ import annotations
import typing

from .length import *
from .string import *


class Element:
    def validate(self, strict=True):
        pass

    def encode(self):
        return bytes()


class Unit(Element):
    def __init__(self):
        pass

    def encode(self):
        return bytes([0])


class Value(Element):
    def __init__(self, ident: int):
        self.ident = ident

    def encode(self):
        return bytes([16 + self.encode_prefix()]) + self.encode_body()

    def encode_prefix(self):
        return self.ident

    def encode_body(self):
        return bytes()


class Option(Element):
    def __init__(self, value: typing.Optional[Element]):
        self.value = value

    @property
    def is_some(self) -> bool:
        return self.value is not None

    @property
    def is_none(self) -> bool:
        return self.value is None

    def __repr__(self):
        return "Option(" + repr(self.value) + ")"

    def encode(self):
        if self.value is not None:
            return bytes([33]) + self.value.encode()
        else:
            return bytes([32])

    def __bool__(self):
        return self.is_some


class Array(Element):
    def __init__(self, data: typing.List[Element]):
        self.value = data

    def encode(self):
        build = bytes([48]) + encode_length(len(self.value))
        for key in self.value:
            build += key.encode()
        return build

    def __repr__(self):
        return "Array(" + repr(self.value) + ")"

    def __iter__(self):
        return self.value.__iter__()


class Structure(Element):
    def __init__(self, value: typing.Dict[str, Element]):
        self.value = value

    def keys(self) -> typing.List[str]:
        return [str(x) for x in self.value.keys()]

    def validate(self, strict=True):
        for key, value in self.value.items():
            if not (type(key) == str):
                raise KeyError(f"Key '{key}' is type {type(key)}, not str")
            if not isinstance(value, Element):
                raise ValueError(f"Value with key '{key}' is type {type(value)}, not Element")

    def __repr__(self):
        return "Structure(" + repr(self.value) + ")"

    def encode(self):
        build = bytes([64]) + encode_length(len(self.value))
        for key, value in self.value.items():
            build += encode_string(key)
            build += value.encode()
        return build

    def __iter__(self):
        return self.value.items().__iter__()


class Variant(Element):
    def __init__(self, name: str, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return "Variant(" + repr(self.name) + ", " + repr(self.value) + ")"

    def encode(self):
        return bytes([80]) + encode_string(self.name) + self.value.encode()


class Map(Element):
    def __init__(self, key_type: typing.Type[Value], data: typing.Dict[Element]):
        self.key_type = key_type
        self.value = data

    def validate(self, strict=True):
        i = 0
        value_type = None
        for key, value in self.value.items():
            if not isinstance(key, Value):
                raise KeyError(f"Key at index {i} is not a valid value ({type(key)})")
            elif not isinstance(key, self.key_type):
                raise KeyError(f"Key at index {i} does not match the parent key type")
            if strict:
                if value_type is None:
                    if not isinstance(value, Element):
                        raise ValueError("Value is not a valid element.")
                    value_type = type(value)
                elif value_type != type(value):
                    raise KeyError(f"Value at index {i} does not match the parent value type")
            i += 1

    def __repr__(self):
        return "Map(" + repr(self.key_type) + ", " + repr(self.value) + ")"

    def encode(self):
        build = bytes([80]) + encode_length(len(self.value))
        for key, value in self.value.items():
            build += key.encode_body()
            build += value.encode()
        return build

    def __iter__(self):
        return self.value.items().__iter__()


class List(Element):
    def __init__(self, item_type: typing.Type[Value], data: typing.List[Value]):
        self.item_type = item_type
        self.value = data

    def validate(self, strict=True):
        i = 0
        for item in self.value:
            if isinstance(item, self.item_type):
                raise ValueError(f"Item at index {i} is not of valid type")
            i += 1

    def encode(self):
        build = bytes([80]) + encode_length(len(self.value))
        for key in self.value:
            build += key.encode_body()
        return build

    def __repr__(self):
        return "List(" + repr(self.item_type) + ", " + repr(self.value) + ")"

    def __iter__(self):
        return self.value.__iter__()



from tycho.elements import *
from tycho.values import *
import typing


def deserialise(s: Element) -> typing.Any:
    if isinstance(s, Unit):
        return None
    elif isinstance(s, Value):
        return s.__getattribute__("value")
    elif isinstance(s, Option):
        if s.is_none:
            return None
        else:
            return deserialise(s)
    elif isinstance(s, Array):
        return [deserialise(e) for e in s.value]
    elif isinstance(s, Structure):
        return {str(k): deserialise(e) for k, e in s.value.items()}
    elif isinstance(s, Variant):
        return dict((s.name, deserialise(s.value)))
    elif isinstance(s, Map):
        return {deserialise_value(k): deserialise(e) for k, e in s.value.items()}
    elif isinstance(s, List):
        return {deserialise_value(k) for k, e in s.value}


def deserialise_value(s: Value):
    if isinstance(s, String) or isinstance(s, Char):
        return s.value
    elif isinstance(s, NumericalValue):
        return s.value
    elif isinstance(s, Boolean):
        return s.value
    elif isinstance(s, Bytes):
        return s.value



from .length import *
from io import BytesIO


def encode_string(s: str):
    data = s.encode("utf-8")
    return encode_length(len(data)) + data


def decode_string(io: BytesIO):
    length = decode_length(io)
    return io.read(length).decode("utf-8")



from .elements import *
from .values import *

from typing import Optional, Type


def serialise(t) -> Element:
    if isinstance(t, dict):
        if is_structure(t):
            return Structure({str(key): serialise(value) for key, value in t.items()})
        map_type = find_array_type(t.keys())
        if map_type is None:
            return Structure({str(key): serialise(value) for key, value in t.items()})
        else:
            return Map(map_type, {serialise_value(key): serialise(value) for key, value in t.items()})
    elif isinstance(t, list) or isinstance(t, tuple):
        list_type = find_array_type(t)
        if list_type is None:
            return Array([serialise(x) for x in list(t)])
        else:
            return List(list_type, [serialise_value(x, typed=list_type) for x in list(t)])
    elif t is None:
        return Option(None)
    else:
        v = serialise_value(t)
        if v is None:
            raise ValueError("Unserializable value parsed")
        else:
            return v


def serialise_value(t, typed=None) -> Optional[Value]:
    if typed is None:
        typed = resolve_value_type(t, num=True)
        if typed is None:
            return None
    return typed(t)


def is_structure(t: dict) -> bool:
    for key in t.keys():
        if type(key) != str:
            return False
    return True


def resolve_value_type(t, num=False) -> Optional[Type[Value]]:
    if isinstance(t, str):
        if len(t) == 1:
            return Char
        else:
            return String
    elif isinstance(t, float):
        return Float64
    elif isinstance(t, bool):
        return Boolean
    elif isinstance(t, int):
        if num is False:
            return NumericalValue
        else:
            return resolve_int_type(t)
    elif isinstance(t, bytes):
        return Bytes
    else:
        return None


def resolve_int_type(t: int) -> Type[Value]:
    if t >= 0:
        if t < 0xFF:
            return Unsigned8
        elif t < 0xFFFF:
            return Unsigned16
        elif t < 0xFFFFFFFF:
            return Unsigned32
        elif t < 0xFFFFFFFFFFFFFFFF:
            return Unsigned64
        else:
            return Unsigned128
    else:
        if t >= -0x8F:
            return Signed8
        elif t >= -0x8FFF:
            return Signed16
        elif t >= -0x8FFFFFFF:
            return Signed32
        elif t >= -0x8FFFFFFFFFFFFFFF:
            return Signed64
        else:
            return Unsigned128


def find_array_type(t) -> Optional[Type[Value]]:
    types = None
    for x in t:
        if types is None:
            types = resolve_value_type(x)
            if types is None:
                return None
        else:
            ts = resolve_value_type(x)
            if types != ts:
                if (issubclass(types, String) or issubclass(types, Char)) and (
                        issubclass(ts, String) or issubclass(ts, Char)):
                    types = String
                    continue
                else:
                    return None

    if issubclass(types, NumericalValue):
        a, b = min(t), max(t)
        if a < 0:
            if (a * -1) > b:
                return resolve_int_type(a)
            else:
                return resolve_int_type(-b)
        else:
            return resolve_int_type(b)
    else:
        return types



from io import BytesIO

from tycho.elements import *
from tycho.values import *


def decode(b: bytes) -> Element:
    data = BytesIO(b)
    return decode_element(data)


def decode_element(data: BytesIO) -> Element:
    prefix = data.read(1)[0]
    element_type = prefix >> 4
    element_data = prefix & 0xF

    if element_type == 0:
        return Unit()

    elif element_type == 1:
        return decode_value(data, element_data)

    elif element_type == 2:
        if element_data == 0x00:
            return Option(None)
        else:
            return Option(decode_element(data))

    elif element_type == 3:
        length = decode_length(data)
        build = []
        for _i in range(0, length):
            build.append(decode_element(data))
        return Array(build)

    elif element_type == 4:
        length = decode_length(data)
        build = {}
        for _i in range(0, length):
            key = decode_string(data)
            value = decode_element(data)
            build[key] = value
        return Structure(build)

    elif element_type == 5:
        name = decode_string(data)
        value = decode_element(data)
        return Variant(name, value)

    elif element_type == 6:
        length = decode_length(data)
        build = {}
        for _i in range(0, length):
            key = decode_value(data, element_data)
            value = decode_element(data)
            build[key] = value
        return Map(get_element_type_from_ident(element_data), build)

    elif element_type == 7:
        length = decode_length(data)
        build = []
        for _ in range(0, length):
            build.append(decode_value(data, element_data))
        return List(get_element_type_from_ident(element_data), build)


def decode_value(data: BytesIO, typed) -> Value:
    if typed == 0:
        return Boolean(data.read(1)[0] == 0x01)
    elif typed == 1:
        return Unsigned8(int.from_bytes(data.read(1), "big", signed=False))
    elif typed == 2:
        return Unsigned16(int.from_bytes(data.read(2), "big", signed=False))
    elif typed == 3:
        return Unsigned32(int.from_bytes(data.read(4), "big", signed=False))
    elif typed == 4:
        return Unsigned64(int.from_bytes(data.read(8), "big", signed=False))
    elif typed == 5:
        return Unsigned128(int.from_bytes(data.read(16), "big", signed=False))
    elif typed == 6:
        return Signed8(int.from_bytes(data.read(1), "big", signed=True))
    elif typed == 7:
        return Signed16(int.from_bytes(data.read(2), "big", signed=True))
    elif typed == 8:
        return Signed32(int.from_bytes(data.read(4), "big", signed=True))
    elif typed == 9:
        return Signed64(int.from_bytes(data.read(8), "big", signed=True))
    elif typed == 10:
        return Signed128(int.from_bytes(data.read(16), "big", signed=True))
    elif typed == 11:
        return Float32(struct.unpack(">f", data.read(4))[0])
    elif typed == 12:
        return Float32(struct.unpack(">d", data.read(8))[0])
    elif typed == 13:
        size = decode_length(data)
        s = data.read(size).decode("utf-8")
        return Char(s[0])
    elif typed == 14:
        return String(decode_string(data))
    elif typed == 15:
        size = decode_length(data)
        return Bytes(data.read(size))
    else:
        raise ValueError(f"Failed to decode value with type ident {typed}")


def get_element_type_from_ident(ident):
    if ident == 0:
        return Boolean
    elif ident == 1:
        return Unsigned8
    elif ident == 2:
        return Unsigned16
    elif ident == 3:
        return Unsigned32
    elif ident == 4:
        return Unsigned64
    elif ident == 5:
        return Unsigned128
    elif ident == 6:
        return Signed8
    elif ident == 7:
        return Signed16
    elif ident == 8:
        return Signed32
    elif ident == 9:
        return Signed64
    elif ident == 10:
        return Signed128
    elif ident == 11:
        return Float32
    elif ident == 12:
        return Float32
    elif ident == 13:
        return Char
    elif ident == 14:
        return String
    elif ident == 15:
        return Bytes
    else:
        raise ValueError(f"Cannot get type from ident {ident}")



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    readme = fh.read()

setuptools.setup(
    name="tycho-py",
    version="0.1.1",
    author="SamHDev",
    author_email="sam.fucked.up@samh.dev",
    description="A python implementation of the Tycho binary format",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/samhdev/tycho-py",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from __future__ import annotations
import typing

from .elements import Value, encode_string
from .length import encode_length
import struct


class Boolean(Value):
    value: bool

    def __init__(self, value: bool):
        super().__init__(0)
        self.value = value

    def __repr__(self):
        return "Boolean(" + repr(self.value) + ")"

    def __str__(self):
        return self.value

    def __bool__(self):
        return self.value

    def encode_body(self):
        return bytes([int(self.value)])


class NumericalValue(Value):
    value: int

    def __init__(self, value: int, _min: int, _max: int, byte_size: int, signed: bool, ident: int):
        super().__init__(ident)
        self.value = value
        self.byte_size = byte_size
        self.signed = signed
        if _min >= value or _max <= value:
            raise ValueError(f"Value {value} is out of bounds ({_min}-{_max})")

    def __repr__(self):
        return self.__class__.__name__ + "(" + repr(self.value) + ")"

    def __str__(self):
        return self.value

    def __bool__(self):
        return self.value

    def __int__(self):
        return self.value

    def __float__(self):
        return float(self.value)

    def encode_body(self):
        return self.value.to_bytes(self.byte_size, "big", signed=self.signed)


class Unsigned8(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, 0, 0xFF, 1, False, 1)


class Unsigned16(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, 0, 0xFFFF, 2, False, 2)


class Unsigned32(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, 0, 0xFFFFFFFF, 4, False, 3)


class Unsigned64(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, 0, 0xFFFFFFFFFFFFFFFF, 8, False, 4)


class Unsigned128(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, 0, 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, 16, False, 5)


class Signed8(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, -0x8F, 0x8F, 1, True, 6)


class Signed16(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, -0x8FFF, 0x8FFF, 2, True, 7)


class Signed32(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, -0x8FFFFFFF, 0x8FFFFFFF, 4, True, 8)


class Signed64(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, -0x8FFFFFFFFFFFFFFFFF, 0x8FFFFFFFFFFFFFFFFF, 8, True, 9)


class Signed128(NumericalValue):
    def __init__(self, value: int):
        super().__init__(value, -0x8FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, 0x8FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, 16, True,
                         10)


class Float32(Value):
    def __init__(self, value: float):
        super().__init__(11)
        self.value = value

    def encode_body(self):
        return struct.pack('>f', self.value)

    def __repr__(self):
        return f"Float32({self.value})"

    def __str__(self):
        return str(self.value)

    def __float__(self):
        return self.value

    def __int__(self):
        return int(self.value)


class Float64(Value):
    def __init__(self, value: float):
        super().__init__(12)
        self.value = value

    def encode_body(self):
        return struct.pack('>d', self.value)

    def __repr__(self):
        return f"Float64({self.value})"

    def __str__(self):
        return str(self.value)

    def __float__(self):
        return self.value

    def __int__(self):
        return int(self.value)


class Char(Value):
    value: str

    def __init__(self, value: str):
        super().__init__(13)
        self.value = value

    def validate(self, strict=True):
        if len(self.value) != 1:
            raise ValueError("Value is not a char (length != 1)")

    def __repr__(self):
        return "String(" + repr(self.value) + ")"

    def __str__(self):
        return self.value

    def __len__(self):
        return len(self.value)

    @property
    def length(self) -> int:
        return len(self.value)

    @property
    def byte_length(self) -> int:
        return len(self.value.encode("utf-8"))

    def encode_body(self):
        data = self.value.encode("utf-8")
        return encode_length(len(data)) + data

    def __iter__(self):
        return self.value.__iter__()


class String(Value):
    value: str

    def __init__(self, value: str):
        super().__init__(14)
        self.value = value

    @property
    def length(self) -> int:
        return len(self.value)

    @property
    def byte_length(self) -> int:
        return len(self.value.encode("utf-8"))

    def __len__(self):
        return len(self.value)

    def __repr__(self):
        return "String(" + repr(self.value) + ")"

    def __str__(self):
        return self.value

    def encode_body(self):
        return encode_string(self.value)

    def __iter__(self):
        return self.value.__iter__()


class Bytes(Value):
    value: bytes

    def __init__(self, value: bytes):
        super().__init__(15)
        self.value = value

    @property
    def length(self) -> int:
        return len(self.value)


    @staticmethod
    def from_list(data: typing.Iterable[int]):
        i = 0
        for byte in data:
            if not (0 <= byte < 256):
                raise ValueError(f"Value at index {i} is not a valid byte. ({byte})")
            i += 1
        return Bytes(bytes(data))

    def __repr__(self):
        return "Bytes(" + str(self.value) + ")"

    def encode_body(self):
        return encode_length(len(self.value)) + self.value

    def __len__(self):
        return len(self.value)

    def __iter__(self):
        return self.value.__iter__()
